# -*- coding: utf-8 -*-
"""
Created on Wed Sep 29 10:44:43 2021
"""
import copy

from sklearn.metrics import roc_auc_score

from utilities import plot_roc, plot_prc

import numpy as np
import torch
import torch.nn as nn
from torch.utils.data import DataLoader

from matplotlib import pyplot as plt

import heterograph_construction

from datetime import datetime

from deepsnap.dataset import GraphDataset
from deepsnap.batch import Batch
from deepsnap.hetero_gnn import (
    HeteroSAGEConv,
    HeteroConv,
    forward_op
)

edges = [('disorder', 'dis_dru_the', 'drug')]
device = 'cuda'


# ---------------------------
# FUNCTIONS
# ---------------------------

def generate_convs_link_pred_layers(hete, conv, hidden_size):
    convs1 = {}
    convs2 = {}
    for message_type in hete.message_types:
        n_type = message_type[0]
        s_type = message_type[2]
        n_feat_dim = hete.num_node_features(n_type)
        s_feat_dim = hete.num_node_features(s_type)
        convs1[message_type] = conv(n_feat_dim, hidden_size, s_feat_dim)
        convs2[message_type] = conv(hidden_size, hidden_size, hidden_size)
    return [convs1, convs2]

class HeteroGNN(torch.nn.Module):
    def __init__(self, convs, hetero, hidden_size):
        super(HeteroGNN, self).__init__()

        self.convs1 = HeteroConv(convs[0])  # Wrap the heterogeneous GNN layers
        self.convs2 = HeteroConv(convs[1])
        self.loss_fn = torch.nn.BCEWithLogitsLoss()
        self.bns1 = nn.ModuleDict()
        self.bns2 = nn.ModuleDict()
        self.relus1 = nn.ModuleDict()
        self.relus2 = nn.ModuleDict()
        self.post_mps = nn.ModuleDict()

        for node_type in hetero.node_types:
            self.bns1[node_type] = torch.nn.BatchNorm1d(hidden_size)
            self.bns2[node_type] = torch.nn.BatchNorm1d(hidden_size)
            self.relus1[node_type] = nn.LeakyReLU()
            self.relus2[node_type] = nn.LeakyReLU()

    def forward(self, data):
        x = data.node_feature
        edge_index = data.edge_index
        x = self.convs1(x, edge_index)
        x = forward_op(x, self.bns1)
        x = forward_op(x, self.relus1)
        x = self.convs2(x, edge_index)
        x = forward_op(x, self.bns2)

        pred = {}
        for message_type in edges:
            nodes_first = None
            nodes_second = None
            if message_type == ('disorder', 'dis_dru_the', 'drug'):
                nodes_first = torch.index_select(x['disorder'], 0, data.edge_label_index[message_type][0, :].long())
                nodes_second = torch.index_select(x['drug'], 0, data.edge_label_index[message_type][1, :].long())

            elif message_type == ('drug', 'dru_dis_the', 'disorder'):
                nodes_first = torch.index_select(x['drug'], 0, data.edge_label_index[message_type][0, :].long())
                nodes_second = torch.index_select(x['disorder'], 0, data.edge_label_index[message_type][1, :].long())

            elif message_type == ('disorder', 'dse_sym', 'disorder'):
                nodes_first = torch.index_select(x['disorder'], 0, data.edge_label_index[message_type][0, :].long())
                nodes_second = torch.index_select(x['disorder'], 0, data.edge_label_index[message_type][1, :].long())

            elif message_type == ('disorder', 'sym_dse', 'disorder'):
                nodes_first = torch.index_select(x['disorder'], 0, data.edge_label_index[message_type][0, :].long())
                nodes_second = torch.index_select(x['disorder'], 0, data.edge_label_index[message_type][1, :].long())

            pred[message_type] = torch.sigmoid(torch.sum(nodes_first * nodes_second, dim=-1))
        return pred

    def predict_all(self, data):
        x = data.node_feature
        edge_index = data.edge_index
        x = self.convs1(x, edge_index)
        x = forward_op(x, self.bns1)
        x = forward_op(x, self.relus1)
        x = self.convs2(x, edge_index)
        x = forward_op(x, self.bns2)

        pred = {}
        for message_type in edges:
            nodes_first = None
            nodes_second = None
            if message_type == ('disorder', 'dis_dru_the', 'drug'):
                nodes_first = x['disorder']
                nodes_second = x['drug']

            elif message_type == ('drug', 'dru_dis_the', 'disorder'):
                nodes_first = x['drug']
                nodes_second = x['disorder']

            elif message_type == ('disorder', 'dse_sym', 'disorder'):
                nodes_first = x['disorder']
                nodes_second = x['disorder']

            elif message_type == ('disorder', 'sym_dse', 'disorder'):
                nodes_first = x['disorder']
                nodes_second = x['disorder']

            for i, elem in enumerate(nodes_first):
                pred[message_type, i] = torch.sigmoid(torch.sum(elem * nodes_second, dim=-1))
        return pred

    def predict_all_type(self, data, type, id):
        x = data.node_feature
        edge_index = data.edge_index
        x = self.convs1(x, edge_index)
        x = forward_op(x, self.bns1)
        x = forward_op(x, self.relus1)
        x = self.convs2(x, edge_index)
        x = forward_op(x, self.bns2)

        pred = {}
        for message_type in edges:
            nodes_first = None
            nodes_second = None
            if message_type == ('disorder', 'dis_dru_the', 'drug') and type == 'disease':
                nodes_first = x['disorder'][id].unsqueeze(0)
                nodes_second = x['drug']

            elif message_type == ('disorder', 'dis_dru_the', 'drug') and type == 'drug':
                nodes_first = x['disorder']
                nodes_second = x['drug'][id].unsqueeze(0)

            elif message_type == ('drug', 'dru_dis_the', 'disorder'):
                nodes_first = x['drug']
                nodes_second = x['disorder']

            elif message_type == ('disorder', 'dse_sym', 'disorder'):
                nodes_first = x['disorder']
                nodes_second = x['disorder']

            elif message_type == ('disorder', 'sym_dse', 'disorder'):
                nodes_first = x['disorder']
                nodes_second = x['disorder']

            for i, elem in enumerate(nodes_first):
                pred[message_type, i] = torch.sigmoid(torch.sum(elem * nodes_second, dim=-1))
        return pred

    def pred(self, data, eid):
        x = data.node_feature
        edge_index = data.edge_index
        x = self.convs1(x, edge_index)
        x = forward_op(x, self.bns1)
        x = forward_op(x, self.relus1)
        x = self.convs2(x, edge_index)
        x = forward_op(x, self.bns2)

        heads = x['disorder']
        tails = x['drug']
        pred = []
        for head, tail in zip(eid[0], eid[1]):
            pred.append(torch.sigmoid(torch.sum(heads[head] * tails[tail], dim=-1)))
        return pred

    def loss(self, pred, y):
        loss = 0
        for key in pred:
            p = pred[key]
            loss += self.loss_fn(p, y[key].type(pred[key].dtype))
        return loss


def train(model, dataloaders, optimizer, epochs):
    val_max = 0
    best_model = model
    t_accu = []
    v_accu = []
    e_accu = []
    lossL = []
    lossLV = []
    for epoch in range(1, epochs + 1):
        for iter_i, batch in enumerate(dataloaders['train']):
            batch.to(device)
            model.train()
            optimizer.zero_grad()
            pred = model(batch)
            loss = model.loss(pred, batch.edge_label)
            if epoch == epochs:
                loss.backward()
            else:
                loss.backward(retain_graph=True)
            optimizer.step()

            log = '     Epoch: {:03d}, Loss: {:.4f}, Train: {:.4f}, Val: {:.4f}, Test: {:.4f}'
            accs = test(model, dataloaders)
            t_accu.append(accs['train'])
            v_accu.append(accs['val'])
            e_accu.append(accs['test'])
            lossL.append(loss.cpu().detach().numpy())
            lossLV.append(accs['valLoss'].cpu().detach().numpy())

            print(log.format(epoch, loss.item(), accs['train'], accs['val'], accs['test']))
            if val_max < accs['val']:
                val_max = accs['val']
                best_model = copy.deepcopy(model)
                
            torch.cuda.empty_cache()

    log = 'Best: Train: {:.4f}, Val: {:.4f}, Test: {:.4f}'
    accs = test(best_model, dataloaders)
    print(log.format(accs['train'], accs['val'], accs['test']))
    
    plt.plot(lossL)
    plt.plot(lossLV)
    plt.legend(['Train', 'Validation'])
    plt.ylabel('Loss')
    plt.xlabel('Iteration')
    plt.xticks(range(0, epochs+1, int(epochs/10)))
    plt.savefig('metrics/lossDMSR-F.svg', format='svg', dpi=1200)
    plt.clf()
    plt.close()
    
    return t_accu, v_accu, e_accu


def test(model, dataloaders):
    model.eval()
    accs = {}
    for mode, dataloader in dataloaders.items():
        acc = 0
        for i, batch in enumerate(dataloader):
            num = 0
            batch.to(device)
            pred = model(batch)
            loss = model.loss(pred, batch.edge_label)
            for key in pred:
                p = pred[key].cpu().detach().numpy()
                pred_label = np.zeros_like(p, dtype=np.int64)
                pred_label[np.where(p > 0.5)[0]] = 1
                pred_label[np.where(p <= 0.5)[0]] = 0
                acc += np.sum(pred_label == batch.edge_label[key].cpu().numpy())
                num += len(pred_label)
        accs[mode] = acc / num
        accs[mode + 'Loss'] = loss
    return accs


def test2(model, test_loader):
    pred_labels = []
    true_labels = []
    keys = []
    pure_pred_labels = []
    for i, batch in enumerate(test_loader):
        batch.to(device)
        pred = model(batch)
        for key in pred:
            p = pred[key].cpu().detach().numpy()
            pred_label = np.zeros_like(p, dtype=np.int64)
            pred_label[np.where(p > 0.5)[0]] = 1
            pred_label[np.where(p <= 0.5)[0]] = 0
            pred_labels.append(pred_label)
            pure_pred_labels.append(p)
            true_labels.append(batch.edge_label[key].cpu().detach().numpy())
            keys.append(key)
    return pred_labels, pure_pred_labels, true_labels, keys


def main(epochs, hidden_dim, lr, weight_decay):

    constructor = heterograph_construction.DISNETConstructor(device=device)
    hetero, _ = constructor.graphWithoutRepoDBInferenceSnap()

    print(hetero.num_edges)
    edge_train_mode = 'disjoint'
    print('edge train mode: {}'.format(edge_train_mode))
    dataset = GraphDataset(
        [hetero],
        task='link_pred',
        edge_train_mode=edge_train_mode,
        edge_message_ratio=0.8,
        edge_negative_sampling_ratio=1
    )

    dataset_train, dataset_val, dataset_test = dataset.split(
        transductive=True, split_ratio=[0.8, 0.1, 0.1], shuffle=True
    )

    train_loader = DataLoader(
        dataset_train, collate_fn=Batch.collate(), batch_size=1
    )
    val_loader = DataLoader(
        dataset_val, collate_fn=Batch.collate(), batch_size=1
    )
    test_loader = DataLoader(
        dataset_test, collate_fn=Batch.collate(), batch_size=1
    )
    dataloaders = {
        'train': train_loader, 'val': val_loader, 'test': test_loader
    }

    convs = generate_convs_link_pred_layers(hetero, HeteroSAGEConv, hidden_dim)
    model = HeteroGNN(convs, hetero, hidden_dim).to(device)
    optimizer = torch.optim.Adam(model.parameters(), lr=lr, weight_decay=weight_decay)

    print("Started training at", datetime.now().strftime("%H:%M:%S"))
    t_accu, v_accu, e_accu = train(model, dataloaders, optimizer, epochs)
    print("Finished training at", datetime.now().strftime("%H:%M:%S"))

    torch.save(model.state_dict(), "models/dmsr")

    # Testing
    model.eval()

    print("Started testing at", datetime.now().strftime("%H:%M:%S"))
    pred_labels, pure_pred_labels, true_labels, keys = test2(model, test_loader)
    print("Finished testing at", datetime.now().strftime("%H:%M:%S"))

    labels = [item for sublist in true_labels for item in sublist]
    predictions = [item for sublist in pred_labels for item in sublist]
    pure_predictions = [item for sublist in pure_pred_labels for item in sublist]

    score = roc_auc_score(labels, pure_predictions)
    print("AUC General:", score)

    plot_roc(labels, pure_predictions, keys[0], "DMSR-F")
    plot_prc(torch.tensor(labels), pure_predictions, keys[0], "DMSR-F")
    return model


if __name__ == '__main__':
    main(400, 32, 0.01, 1e-4)
